/***********************************************************************************

  The Curupira-1 block cipher
  designed by Marcos A. Simplicio Jr. and Paulo S. L. M. Barreto
  Implemented by Alexander Myasnikow
  Web: www.darksoftware.narod.ru

***********************************************************************************/



/************************************************************************************
*                                        INITIAL DEFINES                            *
************************************************************************************/

#include<string.h>
#define rom                     //reserved word in some platforms such as the PIC

#define USE_TABS                //table for S-Box
#define USE_TABX                //table for xtimes operation
#define USE_TABC                //table for ctimes operation


/************************************************************************************
*                                        GLOBAL VARIABLES                           *
************************************************************************************/

//--------------------------- COMPLETE S-BOX ---------------------------

#ifdef USE_TABS
//The 256-byte S-Box table
rom static const unsigned char tabS[256] = {
  0xba, 0x54, 0x2f, 0x74, 0x53, 0xd3, 0xd2, 0x4d,
  0x50, 0xac, 0x8d, 0xbf, 0x70, 0x52, 0x9a, 0x4c,
  0xea, 0xd5, 0x97, 0xd1, 0x33, 0x51, 0x5b, 0xa6,
  0xde, 0x48, 0xa8, 0x99, 0xdb, 0x32, 0xb7, 0xfc,
  0xe3, 0x9e, 0x91, 0x9b, 0xe2, 0xbb, 0x41, 0x6e,
  0xa5, 0xcb, 0x6b, 0x95, 0xa1, 0xf3, 0xb1, 0x02,
  0xcc, 0xc4, 0x1d, 0x14, 0xc3, 0x63, 0xda, 0x5d,
  0x5f, 0xdc, 0x7d, 0xcd, 0x7f, 0x5a, 0x6c, 0x5c,
  0xf7, 0x26, 0xff, 0xed, 0xe8, 0x9d, 0x6f, 0x8e,
  0x19, 0xa0, 0xf0, 0x89, 0x0f, 0x07, 0xaf, 0xfb,
  0x08, 0x15, 0x0d, 0x04, 0x01, 0x64, 0xdf, 0x76,
  0x79, 0xdd, 0x3d, 0x16, 0x3f, 0x37, 0x6d, 0x38,
  0xb9, 0x73, 0xe9, 0x35, 0x55, 0x71, 0x7b, 0x8c,
  0x72, 0x88, 0xf6, 0x2a, 0x3e, 0x5e, 0x27, 0x46,
  0x0c, 0x65, 0x68, 0x61, 0x03, 0xc1, 0x57, 0xd6,
  0xd9, 0x58, 0xd8, 0x66, 0xd7, 0x3a, 0xc8, 0x3c,
  0xfa, 0x96, 0xa7, 0x98, 0xec, 0xb8, 0xc7, 0xae,
  0x69, 0x4b, 0xab, 0xa9, 0x67, 0x0a, 0x47, 0xf2,
  0xb5, 0x22, 0xe5, 0xee, 0xbe, 0x2b, 0x81, 0x12,
  0x83, 0x1b, 0x0e, 0x23, 0xf5, 0x45, 0x21, 0xce,
  0x49, 0x2c, 0xf9, 0xe6, 0xb6, 0x28, 0x17, 0x82,
  0x1a, 0x8b, 0xfe, 0x8a, 0x09, 0xc9, 0x87, 0x4e,
  0xe1, 0x2e, 0xe4, 0xe0, 0xeb, 0x90, 0xa4, 0x1e,
  0x85, 0x60, 0x00, 0x25, 0xf4, 0xf1, 0x94, 0x0b,
  0xe7, 0x75, 0xef, 0x34, 0x31, 0xd4, 0xd0, 0x86,
  0x7e, 0xad, 0xfd, 0x29, 0x30, 0x3b, 0x9f, 0xf8,
  0xc6, 0x13, 0x06, 0x05, 0xc5, 0x11, 0x77, 0x7c,
  0x7a, 0x78, 0x36, 0x1c, 0x39, 0x59, 0x18, 0x56,
  0xb3, 0xb0, 0x24, 0x20, 0xb2, 0x92, 0xa3, 0xc0,
  0x44, 0x62, 0x10, 0xb4, 0x84, 0x43, 0x93, 0xc2,
  0x4a, 0xbd, 0x8f, 0x2d, 0xbc, 0x9c, 0x6a, 0x40,
  0xcf, 0xa2, 0x80, 0x4f, 0x1f, 0xca, 0xaa, 0x42,
};

//The S-Box function uses the complete table
#define sBox(value) (tabS[value])

//--------------------------- MINI S-BOXES ---------------------------
#else

static const unsigned char P[16] = {
  0x3, 0xF, 0xE, 0x0, 0x5, 0x4, 0xB, 0xC, 0xD, 0xA, 0x9, 0x6, 0x7, 0x8, 0x2,
    0x1
};
static const unsigned char Q[16] = {
  0x9, 0xE, 0x5, 0x6, 0xA, 0x2, 0x3, 0xC, 0xF, 0x0, 0x4, 0xD, 0x7, 0xB, 0x1,
    0x8
};

static unsigned char
sBox (const unsigned char u)    // The S-Box calculation
{
  static unsigned char uHigh, uLow, tempU;

  //First pass
  uHigh = P[(u >> 4)];
  uLow = Q[u & 0xF];

  //Second Pass
  tempU = Q[(uHigh & 0xC) ^ ((uLow >> 2))];
  uLow = P[((uHigh << 2) & 0xC) ^ (uLow & 0x3)];

  //Third pass
  uHigh = P[(tempU & 0xC) ^ ((uLow >> 2))];
  uLow = Q[((tempU << 2) & 0xC) ^ (uLow & 0x3)];

  //Concatenation
  return (uHigh << 4) ^ uLow;
}

#endif

//-------------- X-TIMES OPPERATION --------------
#ifdef USE_TABX
rom static const unsigned char tabX[256] = {
  0x00, 0x02, 0x04, 0x06, 0x08, 0x0A, 0x0C, 0x0E,
  0x10, 0x12, 0x14, 0x16, 0x18, 0x1A, 0x1C, 0x1E,
  0x20, 0x22, 0x24, 0x26, 0x28, 0x2A, 0x2C, 0x2E,
  0x30, 0x32, 0x34, 0x36, 0x38, 0x3A, 0x3C, 0x3E,
  0x40, 0x42, 0x44, 0x46, 0x48, 0x4A, 0x4C, 0x4E,
  0x50, 0x52, 0x54, 0x56, 0x58, 0x5A, 0x5C, 0x5E,
  0x60, 0x62, 0x64, 0x66, 0x68, 0x6A, 0x6C, 0x6E,
  0x70, 0x72, 0x74, 0x76, 0x78, 0x7A, 0x7C, 0x7E,
  0x80, 0x82, 0x84, 0x86, 0x88, 0x8A, 0x8C, 0x8E,
  0x90, 0x92, 0x94, 0x96, 0x98, 0x9A, 0x9C, 0x9E,
  0xA0, 0xA2, 0xA4, 0xA6, 0xA8, 0xAA, 0xAC, 0xAE,
  0xB0, 0xB2, 0xB4, 0xB6, 0xB8, 0xBA, 0xBC, 0xBE,
  0xC0, 0xC2, 0xC4, 0xC6, 0xC8, 0xCA, 0xCC, 0xCE,
  0xD0, 0xD2, 0xD4, 0xD6, 0xD8, 0xDA, 0xDC, 0xDE,
  0xE0, 0xE2, 0xE4, 0xE6, 0xE8, 0xEA, 0xEC, 0xEE,
  0xF0, 0xF2, 0xF4, 0xF6, 0xF8, 0xFA, 0xFC, 0xFE,
  0x4D, 0x4F, 0x49, 0x4B, 0x45, 0x47, 0x41, 0x43,
  0x5D, 0x5F, 0x59, 0x5B, 0x55, 0x57, 0x51, 0x53,
  0x6D, 0x6F, 0x69, 0x6B, 0x65, 0x67, 0x61, 0x63,
  0x7D, 0x7F, 0x79, 0x7B, 0x75, 0x77, 0x71, 0x73,
  0x0D, 0x0F, 0x09, 0x0B, 0x05, 0x07, 0x01, 0x03,
  0x1D, 0x1F, 0x19, 0x1B, 0x15, 0x17, 0x11, 0x13,
  0x2D, 0x2F, 0x29, 0x2B, 0x25, 0x27, 0x21, 0x23,
  0x3D, 0x3F, 0x39, 0x3B, 0x35, 0x37, 0x31, 0x33,
  0xCD, 0xCF, 0xC9, 0xCB, 0xC5, 0xC7, 0xC1, 0xC3,
  0xDD, 0xDF, 0xD9, 0xDB, 0xD5, 0xD7, 0xD1, 0xD3,
  0xED, 0xEF, 0xE9, 0xEB, 0xE5, 0xE7, 0xE1, 0xE3,
  0xFD, 0xFF, 0xF9, 0xFB, 0xF5, 0xF7, 0xF1, 0xF3,
  0x8D, 0x8F, 0x89, 0x8B, 0x85, 0x87, 0x81, 0x83,
  0x9D, 0x9F, 0x99, 0x9B, 0x95, 0x97, 0x91, 0x93,
  0xAD, 0xAF, 0xA9, 0xAB, 0xA5, 0xA7, 0xA1, 0xA3,
  0xBD, 0xBF, 0xB9, 0xBB, 0xB5, 0xB7, 0xB1, 0xB3,
};

#define xTimes(value) (tabX[value])

#else
//#define xTimes(value) (((value) & 0x80) ? ((value) << 1) ^ 0x4D : ((value) << 1))
static unsigned char
xTimes (const unsigned char value)
{
  if (value & 0x80)
    return ((value << 1) ^ 0x4D);
  else
    return (value << 1);
}
#endif

#ifdef USE_TABC
rom static const unsigned char tabC[256] = {
  0x00, 0x1C, 0x38, 0x24, 0x70, 0x6C, 0x48, 0x54,
  0xE0, 0xFC, 0xD8, 0xC4, 0x90, 0x8C, 0xA8, 0xB4,
  0x8D, 0x91, 0xB5, 0xA9, 0xFD, 0xE1, 0xC5, 0xD9,
  0x6D, 0x71, 0x55, 0x49, 0x1D, 0x01, 0x25, 0x39,
  0x57, 0x4B, 0x6F, 0x73, 0x27, 0x3B, 0x1F, 0x03,
  0xB7, 0xAB, 0x8F, 0x93, 0xC7, 0xDB, 0xFF, 0xE3,
  0xDA, 0xC6, 0xE2, 0xFE, 0xAA, 0xB6, 0x92, 0x8E,
  0x3A, 0x26, 0x02, 0x1E, 0x4A, 0x56, 0x72, 0x6E,
  0xAE, 0xB2, 0x96, 0x8A, 0xDE, 0xC2, 0xE6, 0xFA,
  0x4E, 0x52, 0x76, 0x6A, 0x3E, 0x22, 0x06, 0x1A,
  0x23, 0x3F, 0x1B, 0x07, 0x53, 0x4F, 0x6B, 0x77,
  0xC3, 0xDF, 0xFB, 0xE7, 0xB3, 0xAF, 0x8B, 0x97,
  0xF9, 0xE5, 0xC1, 0xDD, 0x89, 0x95, 0xB1, 0xAD,
  0x19, 0x05, 0x21, 0x3D, 0x69, 0x75, 0x51, 0x4D,
  0x74, 0x68, 0x4C, 0x50, 0x04, 0x18, 0x3C, 0x20,
  0x94, 0x88, 0xAC, 0xB0, 0xE4, 0xF8, 0xDC, 0xC0,
  0x11, 0x0D, 0x29, 0x35, 0x61, 0x7D, 0x59, 0x45,
  0xF1, 0xED, 0xC9, 0xD5, 0x81, 0x9D, 0xB9, 0xA5,
  0x9C, 0x80, 0xA4, 0xB8, 0xEC, 0xF0, 0xD4, 0xC8,
  0x7C, 0x60, 0x44, 0x58, 0x0C, 0x10, 0x34, 0x28,
  0x46, 0x5A, 0x7E, 0x62, 0x36, 0x2A, 0x0E, 0x12,
  0xA6, 0xBA, 0x9E, 0x82, 0xD6, 0xCA, 0xEE, 0xF2,
  0xCB, 0xD7, 0xF3, 0xEF, 0xBB, 0xA7, 0x83, 0x9F,
  0x2B, 0x37, 0x13, 0x0F, 0x5B, 0x47, 0x63, 0x7F,
  0xBF, 0xA3, 0x87, 0x9B, 0xCF, 0xD3, 0xF7, 0xEB,
  0x5F, 0x43, 0x67, 0x7B, 0x2F, 0x33, 0x17, 0x0B,
  0x32, 0x2E, 0x0A, 0x16, 0x42, 0x5E, 0x7A, 0x66,
  0xD2, 0xCE, 0xEA, 0xF6, 0xA2, 0xBE, 0x9A, 0x86,
  0xE8, 0xF4, 0xD0, 0xCC, 0x98, 0x84, 0xA0, 0xBC,
  0x08, 0x14, 0x30, 0x2C, 0x78, 0x64, 0x40, 0x5C,
  0x65, 0x79, 0x5D, 0x41, 0x15, 0x09, 0x2D, 0x31,
  0x85, 0x99, 0xBD, 0xA1, 0xF5, 0xE9, 0xCD, 0xD1,
};
#endif


//-------------- SCHEDULE RESETING CONSTANTS --------------

//Constant for 192-bit keys
rom static const unsigned char STR_CONST_192[24] =
  { 0xFA, 0x89, 0x29, 0xCA, 0x4B, 0x91, 0x58, 0x04,
  0xCA, 0x44, 0xB3, 0x7C, 0x58, 0x19, 0xFA, 0x90,
  0x9A, 0x0B, 0x75, 0x79, 0xF4, 0x4A, 0x5A, 0xDE
};


//-------------- "ATRIBUTES"  --------------

//The cipher current key, its end and the first position in its last column, respectevelly
static unsigned char *key;
static unsigned char *keyEnd;
static unsigned char *ptrLastCol;

//The cipher current block and its end
static unsigned char *block;
static unsigned char *blockEnd;

//The array used to reset the key
const static unsigned char *keyReseter;

//Offset indicating the start position of the constants to be taken by the key schedule
static unsigned char constantsOffSet;

//Direction of the crypt opperation (1:encrypt; 2:decrypt)
static unsigned char direction, compl;

static unsigned char aux1, aux2, aux3;  //auxiliary variables used to store various intermediary values
static unsigned char *ptr1, *ptr2;      //auxiliary pointers used to
static unsigned char i, r;      //used as loop counters



/************************************************************************************
*                              CRYPT FUNCTIONS                                     *
***********************************************************************************/


void __stdcall __export
setup (unsigned char *newKey)
{

  keyReseter = STR_CONST_192;
  constantsOffSet = 0;          //initialize the constants offset (first constant will be taken)
  key = newKey;                 //sets the new key
  keyEnd = newKey + 24;         //points to the position after the end of the key
  ptrLastCol = keyEnd - 3;      //points to the element in the first row and last column of the key
}

/**
   * Creates the key for the next round, replacing the old key.
   * It shall be used  both for encryption and decryption, since the correct parameter
   * "direction" indicates the direction of the scheduling.
   *
   * @param resetFlag Indicates if the schedule shall be reset (the reset constants will be     applied)
   */
static void
createNextKey (const unsigned char resetFlag)
{

  static unsigned char jumper;
  jumper = 1;                   //the jump hasn't yet ocurred (see comments below)

  //the code included in this 'if' statement (xiShift and key-reseting) shall
  //be performed during both encryption and decryption, in diferent times (respectivelly
  //before and after the multiplication by E[^(-1)]). The 'goto' statement at the end of
  //this function, along with the correct specification of the jumper variable are used
  //in order to achieve this result: the 'if' statemente below will run at the *beginning*
  //of the function calling if direction == 1 (encryption) and at the *end* of the function
  //if direction == 2 (decryption)
label:
  if (direction == jumper)
    {

/////////////////////////////////////// XI SHIFT ///////////////////////////////////////

      //Update the key shcedule constant position to be used
      if (direction & 0x02)     //inverse: pre-decrement
        {
          constantsOffSet -= 8;
        }
      aux3 = constantsOffSet;
      if (direction & 0x01)     //direct: pos-increment
        {
          constantsOffSet += 8;
        }

      //Initialization of some important positions and values
      ptr1 = key;               //points to the start of the key
      aux1 = key[direction];    //value of the 1st element to be left shifted

      ptr2 = ptrLastCol + compl;        //points to the first element to be right shifted
      aux2 = *ptr2;             //value of the 1st element to be right shifted

      //Executes the shift and applies constants
      for (i = 1;; ptr1 += compl, ptr2 -= 3)
        {
          *ptr1 ^= sBox (aux3); //apply the schedule constant to 1st row
          ptr1 += direction;    //updates 'ptr1', so it goes to the position to be left shifted
          if (i++ == 8)         //Last column achieved: circular shift shall be done
            {
              break;
            }
          aux3++;
          *ptr1 = *(ptr1 + 3);  //left shift
          *ptr2 = *(ptr2 - 3);  //right shift
        }

      *ptr1 = aux1;
      *ptr2 = aux2;


/////////////////////////////////END OF XI SHIFT////////////////////////////////////////////////////////

    }
  else if (resetFlag)           //The accumulated constants shall be applied
    {

/////////////////////////////////////// KEY RESETING ////////////////////////////////////////////////////
      //resets the constants position accordingly
      constantsOffSet = 0;
      if (direction & 0x02)
        constantsOffSet = 192;

      //Apply the constants for key-reseting
      for (i = 0; i != 24; i++)
        {
          key[i] ^= keyReseter[i];
        }


///////////////////////////END OF KEY RESETING/////////////////////////////////////////
    }

  if (jumper & 0x02)
    return;                     //the jump has already been performed: end of the function

///////////////////MULTIPLICATION BY MATRIX E (or E^(-1))/////////////////////////
  for (i = 0; i != 24;)
    {
      aux1 = key[i] ^ key[i + 1] ^ key[i + 2];

#ifdef USE_TABC
      aux2 = tabC[aux1];
#else
      aux2 = xTimes (xTimes (xTimes (xTimes (aux1) ^ aux1) ^ aux1));    //multiplication by c (= 0x1C)
#endif

      //Multiplication by E^(-1)
      if (direction & 0x02)
        {
          aux2 ^= aux1;
        }

      key[i++] ^= aux2;
      key[i++] ^= aux2;
      key[i++] ^= aux2;
    }
////////////////END OF MULTIPLICATION BY MATRIX E (or E^(-1))////////////////////

  //the jump for the beggining of the function is set to be done (==2)
  jumper++;

  //jumps to the beggining of the function in order to check if the xiShift and key
  //reseting still must be performed
  goto label;

}                               //end of function createNextKey


//Swaps positions given by ptr1 and ptr2 and applies the SBox to them
//(these pointers should be already pointing to ciphertext)
static void
swapCT (void)
{
  aux1 = sBox (*ptr1);
  *ptr1 = sBox (*ptr2);
  *ptr2 = aux1;
}


//Applies the key to one column of the ciphertext and updates ptr1 and ptr2,
//which will point to the next unchanged position of the block and key, respectivelly
static void
applyKey (void)
{
  *ptr1++ ^= sBox (*ptr2);
  ptr2++;
  *ptr1++ ^= *ptr2++;
  *ptr1++ ^= *ptr2++;
}

//Applies the sBox on the first row and them swaps the positions below them (second row)
static void
sOnRow1 (void)
{
  *ptr1++ = sBox (*ptr1);
  *ptr2++ = sBox (*ptr2);
  swapCT ();
}

//Multiplies one row by matrix D
static void
applyD (void)
{
  aux3 = sBox (*ptr2);          //stores the S-Box for the element in the first row, which will be used twice
  ptr2++;
  aux1 = *ptr1 ^ *(ptr1 + 1) ^ *(ptr1 + 2);
  if (direction & 0x02)
    aux1 ^= aux3 ^ *ptr2 ^ *(ptr2 + 1); //the inverse key generation uses the key multiplied by the D matrix
  aux1 = xTimes (aux1);
  aux2 = xTimes (aux1);
  *ptr1++ ^= aux1 ^ aux3;
  *ptr1++ ^= aux2 ^ *ptr2++;
  *ptr1++ ^= aux1 ^ aux2 ^ *ptr2++;
}


/**
   * Encrypt or decrypt the block passed as parameter. The plaintext and the ciphertext arrays
   * are always the same.
   *
   * @param b The [plain/cipher]text array
   * @param dir '1' for encryption; '2' for decryption
   * @param numberOfRound The number of rounds to be adopted
   */
void
c_crypt (unsigned char *b, unsigned char dir, unsigned char numberOfRounds)
{

  //Difference between the maximum number of rounds and the number adopted
  //It's used in order to skip some round keys
  static unsigned char diff;

  //Points to the block and its end
  block = b;
  blockEnd = b + 12;

  //Gets the passed parameters and store them locally
  direction = dir;

  //constant value: complement of direction (modulo 3)
  compl = 0x03 ^ dir;

  //calculates the value of 'diff'
  diff = 1;

  //Go to the first key to be used by the decryption algorithm
  if (dir & 0x02)
    {
      //Calculate the key for the last round permitted (the reseting constants must be applied)
      createNextKey (1);

      //Pass by the keys that won`t be used (only activated when maximum round number wasn't adopted)
      for (i = 1; i < diff; i++)
        createNextKey (0);
    }

  //whitenning
  ptr1 = block;
  ptr2 = key;
  applyKey ();
  applyKey ();
  applyKey ();
  applyKey ();

  //round function applied N-1 times
  for (r = 1;;)
    {

      //Permutation (Pi) layer
      ptr1 = block;
      ptr2 = block + 3;
      sOnRow1 ();               //Applies the Sbox on (0) (3); Swap(1, 4)
      ptr1++;
      ptr2 += 4;
      swapCT ();                //Swap(2, 8)
      ptr1 += 4;
      ptr2++;
      sOnRow1 ();               //Applies the Sbox on (6, 9); Swap(7,10)
      ptr1 -= 2;
      ptr2++;
      swapCT ();                //Swap(5,11)


      //Create the key for this round
      createNextKey (0);

      ptr1 = block;
      ptr2 = key;

      //Checks if the final round is achieved
      if (r++ == numberOfRounds)
        {
          applyKey ();
          applyKey ();
          applyKey ();
          applyKey ();          //Last key addition
          break;
        }

      //Multiplication by D (Theta Layer) and key application
      applyD ();
      applyD ();
      applyD ();
      applyD ();



    }

  //Resets the key (the encryption algorithm doesn't automatically reset it)
  if (direction & 0x01)
    {

      //Passes by the keys that haven`t been used (maximum round number wasn't adopted)
      for (i = 1; i < diff; i++)
        createNextKey (0);

      //Applies the reseting constants to recover the original key
      createNextKey (1);
    }

  return;
}




void __stdcall __export
crypt (unsigned char *b)
{
  c_crypt (&b[0], 1, 23);
}

void __stdcall __export
decrypt (unsigned char *b)
{
  c_crypt (&b[0], 2, 23);
}





typedef unsigned char u8;
typedef unsigned long u32;

u32 __stdcall __export
getblocksize ()
{
  return 96;
}

u32 __stdcall __export
getkeysize ()
{
  return 192;
}

void __stdcall __export
getciphername (u8 * p)
{
  strcpy (p, "Curupira1");
}
